/************************************************************************/
/*	Praca domowa z ZPR
 *	Wzorzec projektowy MONITOR
 *	@autor Piotr Walczak grupa H1ISI
 *
 *	Opis : Jest to wzorzec synchronizacyjny, ktrego zastosowanie ma chroni
 *	dane przed niekontrolowanymi zmianami zwizanymi z dziaaniem na tych 
 *	danych wielu wspbienych wtkw. Zastosowanie wzorca sprowadza si do 
 *	stworzenia dla obiektu synchronizowanego odpowiedniego mechanizmu synchronizacji
 *	w taki sposb, aby by on nie widoczny dla uytkownika tego obiektu.
 *	Zastosowanie monitora daje nam pewno, e w jego obrbie wykonywane jest tylko 
 *	i wycznie jedna metoda w danym czasie (aden wtek nie moe rozpocz wykonywania
 *	metody danego monitora dopuki inny wtek jest w trakcie wykonywania
 *	jakiej metody tego monitora). Zapobiega to wystpieniu wycigw do
 *	zasobw wspdzielonych w zakresie jednego monitora.
 */
/************************************************************************/

#include <iostream>

#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/bind.hpp>
#include <boost/thread/condition.hpp>

using namespace boost;
using namespace std;

/**	Interfejs dla klas mogcych by obsugiwanymi przez clase Monitor	*/
class InterfaceClass{
public :
	virtual void method1() = 0;
	virtual void method2() = 0;
	virtual void print() const = 0;
};

/**	Przykadowa klasa implementujca interfejs InterfaceClass	*/
class SynchroClass : public InterfaceClass{
private :
	/**	Zmienne wykorzystywane do sprawdzenia poprawnoci synchronizacji	*/
	volatile int i_;
	volatile int j_;
public :
	SynchroClass(){
		i_ = 100;
		j_ = 100;
	}
	SynchroClass(const SynchroClass& second){
		i_ = second.i_;
		j_ = second.j_;
	}
	virtual void method1(){
		j_ -= 1;
		i_ += 1;
	}
	virtual void method2(){
		j_ += 1;
		i_ -= 1;
	}

	/**	
	 *	Funkcja wypisuje stan licznikw oraz zawiesza prace programu wypisujc komunikat
	 *	o bdzie w przypadku gdy doszo do wycigu 
	 */
	void print() const {
		cout << "i : " << i_ << " j : " << j_ << " suma : " << i_ + j_ << endl;
		if ( i_ + j_ != 200 ){
			cout << "ERROR :: WYSCIG" << endl;
			system("PAUSE");
		}
	}
};

/**
 *	Monitor z wykorzystaniem agregacji obiektu araz podziaem metod na dwie grupy oddzielnie
 *	synchronizowanych funkcji. Dzieki temu na jednym obiekcie moe pracowa 2 wtki rwnoczenie
 *	o ile pracuj na niezalenych danych np. na dwch rnych obiektach agregowanych przez
 *	monitor.
 */
class Monitor{
private :
	/**	
	 *	Wykorzystanie interfejsu zwiksza funkcjonalno Monitora - aby obiekt mgby 
	 *	synchronizowany przez ten monitor musi jedynie implementowa okrelony interfejs
	 */
	InterfaceClass* const obj_;
	InterfaceClass* const obj2_;

	/**	
	 *	Mutex pierwszej grupy synchronizowanych metod	
	 *	- wykorzystanie wskaznikw pozwala na synchronizacje wielu sekcji krytycznych zawartych w rznych
	 *	monitorach tym samym mutexem
	 */
	boost::mutex* const mutex_;

	/**	Mutex drugiej grupy synchronizowanych metod	*/
	boost::mutex* const mutex2_;
public :
	/**
	 *	Konstruktor monitora
	 *	 - dziki temu, e pobieramy mutexy z zewntrz moliwe jest rozszerzenie funkcjonalnoci
	 *		monitora - wicej ni jeden obiekt (monitor) moe by synchronizowany przy pomocy
	 *		jednego, tego samego mutexu; ponad to rozwizuemy problem z niemoliwoci kopiowania
	 *		mutexw
	 *	@param obj	-	pierwszy objekt podlegajcy synchronizacji
	 *	@param obj2	-	drugi objekt podlegajcy synchronizacji
	 *	@param mutex		-	mutex odpowiedzialny za synchronizowanie pierwszego obiektu
	 *	@param mutex2		-	mutex odpowiedzialny za synchronizowanie drugiego obiektu
	 */
	Monitor(InterfaceClass* obj, InterfaceClass* obj2, boost::mutex *mutex, boost::mutex *mutex2)
		: obj_(obj), obj2_(obj2), mutex_(mutex), mutex2_(mutex2){
		// Do nothing
	}

	Monitor(const Monitor& monitor)
		: obj_(monitor.obj_), obj2_(monitor.obj2_), mutex_(monitor.mutex_), mutex2_(monitor.mutex2_){
		// Do nothing
	}
	/************************************************************************/
	/* Tutej delegujemy metody obiektu InterfaceClass                       */
	/************************************************************************/

	void method1(){
		/**	Opuszczenie mutexu	*/
		boost::mutex::scoped_lock  lock(*mutex_);
		obj_->method1();
	}

	void method2(){
		boost::mutex::scoped_lock  lock(*mutex_);
		obj_->method2();
	}

	void method3(){
		boost::mutex::scoped_lock  lock(*mutex2_);
		obj2_->method1();
	}

	void method4(){
		boost::mutex::scoped_lock  lock(*mutex2_);
		obj2_->method2();
	}

	void print() const {
		boost::mutex::scoped_lock lock(*mutex_);
		boost::mutex::scoped_lock lock2(*mutex2_);
		cout << "objekt 1 : ";
		obj_->print();
		cout << "objekt 2 : ";
		obj2_->print();
	}
};

/**
 *	Wykorzystanie dziedziczenia do stworzenia monitora pozwala wykorzysta polimorfizm
 *	- urzywa monitora dokadnie wszdzie tam gdzie urzywano klasy bazowej
 *	Oczywicie o ile kod wykorzystyjcy klase bazow uwzgldni moliwo wykorzystywania
 *	klas pochodnych do tej klasy bozowej
 */
class MonitorSynchroClass : public SynchroClass {
private :
	/**	Mutex synchronizujcy przecione metody danego obiektu	*/
	boost::mutex *mutex_;
public :
	MonitorSynchroClass(boost::mutex* mutex) 
		: mutex_(mutex){
		//Do nothing
	}

	/************************************************************************/
	/* Przecienie interesujcych nas metod z klasy bazowej                */
	/************************************************************************/
	void method1(){
		boost::mutex::scoped_lock  lock(*mutex_);
		SynchroClass::method1();
	}
	void method2(){
		boost::mutex::scoped_lock  lock(*mutex_);
		SynchroClass::method2();
	}
	void print() const {
		boost::mutex::scoped_lock  lock(*mutex_);
		SynchroClass::print();
	}
};

/************************************************************************/
/* Funkcje reprezentujce kolejne wtki                                 */
/************************************************************************/
void Thread1(Monitor* monitor){
	while( true ){
		monitor->method1();
	}
}

void Thread2(Monitor* monitor){
	while ( true ){
		monitor->method2();
	}
}

void Thread13(Monitor* monitor){
	while( true ){
		monitor->method3();
	}
}

void Thread14(Monitor* monitor){
	while ( true ){
		monitor->method4();
	}
}

void Thread5(MonitorSynchroClass* monitor){
	while( true ){
		monitor->method1();
	}
}

void Thread6(MonitorSynchroClass* monitor){
	while ( true ){
		monitor->method2();
	}
}

int main (){
	SynchroClass cl;
	SynchroClass c2;
	boost::mutex mutex;
	boost::mutex mutex2;
	Monitor monitor( &cl, &c2, &mutex, &mutex2 );

	/**	monitor musi byprzekazany przyz wskaznik gdy bind przekazuje element przez warto	*/
	boost::thread thrd1(boost::bind(&Thread1, &monitor));
	boost::thread thrd2(boost::bind(&Thread2, &monitor));

	boost::thread thrd3(boost::bind(&Thread13, &monitor));
	boost::thread thrd4(boost::bind(&Thread14, &monitor));

	/**	podanie tego samego mutexu co w monitorze monitor - zapewnia synchronizacje wydrukw kontrolnych	*/
	MonitorSynchroClass monitorSynchro(&mutex2);

	boost::thread thrd5(boost::bind(&Thread5, &monitorSynchro));
	boost::thread thrd6(boost::bind(&Thread6, &monitorSynchro));

	while (true){
		monitor.print();
		monitorSynchro.print();
	}

	thrd1.join();
	thrd2.join();
	thrd3.join();
	thrd4.join();
	thrd5.join();
	thrd6.join();

	return 0;
}
